/* symtab.c 
 * Simple symbol table
 *
 * The symbol table is a 1-dim char array. The symbols
 * are fixed length (simulation of 2-dim in 1-dim)
 * For good C compatibility terminate sub-strings with '\0'
 * symtab has several sections:
 * keyword section
 * global section for global vars and func names
 * local section for formal parameters and local vars
 *
 * Search with keywords are reserved words
 * first keywords, then locals, then globals
 *
 * by Andre Adrian
 *
 * version 29may2006
 *
 * fixed length tbd
 */

#include <string.h>
#include <stdio.h>
 
#define SYMTABLEN 1000

char symtab[SYMTABLEN];
int section[4];         
int namespace;
 
void error(char msg[]) {
  fprintf(stderr, msg);
  fprintf(stderr, "\n");
  exit(1);
}
  
 
int lookup(char name[], int namespace) {
// look up name in symbol table section namespace. 
// If found return symbol id in namespace else -1
  char cn, cs;
  int in, is;
  int found, sym;

  in = 0;  cn = name[in];
  is = section[namespace-1];  cs = symtab[is];
  found = 1; sym = 0;
  while (is < section[namespace]) {	// section end
    if (cn != cs) {
      found = 0;
    }
    is = is + 1; cs = symtab[is];
    if (cn != '\0') {
      in = in + 1; cn = name[in];
    }
    if (cs == ' ') {                    // symbol end
      if (found == 1 && cn == '\0') {
      	return sym;
      } else {
	in = 0;  cn = name[in];
	is = is + 1; cs = symtab[is];
	sym = sym + 1;
	found = 1;
      }
    }
  }
  return -1;
}

/* lookupall: suche in keywords, locals, globals */

void addsym(char name[]) {
  if (strlen(symtab) + strlen(name) + 2 <= SYMTABLEN) {
    strcat(symtab, name);
    strcat(symtab, " ");
  } else {
    error("symtab full");
  }
}

int newname(char name[], int namespace) {
// namespace: 1 = keywords, 2 = globals, 3 = locals
  int ident;
  
  ident = lookup(name, namespace);
  if (ident >= 0) {
    return ident;
  } else {
    addsym(name);
    section[namespace] = strlen(symtab);
    return lookup(name, namespace);
  }
}

void getname(char name[], int ident, int namespace) {  
  int i, j;
  
  i = section[namespace-1];
  while (ident > 0) {
    if (' ' == symtab[i]) {
      ident = ident -1;
    }
    i = i + 1;
  }
  j = 0;
  while(symtab[i] != ' ') {
    name[j] = symtab[i];
    j = j + 1;
    i = i + 1;
  }
  name[j] = '\0';
}

char *gn(int ident) {
  static char name[10];
  
  getname(name, ident, namespace);
  return name;
}
         
void initsymtab() {
  strcpy(symtab, "if then else endif while do break wend proc func return int char ");
  section[0] = 0;
  section[1] = strlen(symtab);
  section[2] = section[1];
  section[3] = section[1];
}

void addglobals() {
  namespace = 2;
  section[3] = section[2];  /* delete old locals */
}

void addlocals() {
  namespace = 3;
}

#ifdef MODULTEST

int main(int argc, char *argv[]) {
  int sym;
  initsymtab();
  char name[10];
  
  sym = lookup("if", 1);  
  printf("lookup if =%d\n", sym);
  getname(name, sym, 1);
  printf("getname =%s\n", name);
  
  sym = lookup("iff", 1);  
  printf("lookup iff =%d\n", sym);
  
  sym = lookup("then", 1);    
  printf("lookup then =%d\n", sym);
  getname(name, sym, 1);
  printf("getname =%s\n", name);
  
  sym = lookup("char", 1);    
  printf("lookup char =%d\n", sym);
  getname(name, sym, 1);
  printf("getname =%s\n", name);
  
  sym = lookup("xxx", 1);  
  printf("lookup xxx =%d\n", sym);
  sym = lookup("yyyyyyyyyyy", 1);  
  printf("lookup yyyyyyyyyyy =%d\n", sym);
  
  sym = lookup("vara", 2);  
  printf("lookup vara =%d\n", sym);
  newname("vara", 2);
  sym = lookup("vara", 2);  
  printf("lookup vara =%d\n", sym);
  getname(name, sym, 2);
  printf("getname =%s\n", name);
  
  sym = lookup("varb", 2);  
  printf("lookup varb =%d\n", sym);
  newname("varb", 2);
  sym = lookup("varb", 2);  
  printf("lookup varb =%d\n", sym);
  getname(name, sym, 2);
  printf("getname =%s\n", name);
  
  newname("vara", 2);
}

#endif
